module BackendLlvm where

import GCode
import Control.Monad
import Control.Monad.State
import Data.List(intersperse)
import Data.Char(ord)

data TrState = TrState {
    outcode :: [String],
    snum :: Int,
    vstack :: [Maybe String],
    vsdelta :: Int
}
type TrM = State TrState

data BtTypes = 
    BtUnknown | BtInt | BtInd | BtFloat | BtBool | BtAp | BtHole | BtChar

btnode BtInt = "%.INTNODE"
btnode BtFloat = "%.FLOATNODE"
btnode BtChar = "%.CHARNODE"
btnode BtInd = "%.INDNODE"
btnode BtBool = "%.BOOLNODE"
btnode BtAp = "%.APNODE"
btnode BtHole = "%.HOLENODE"
btnode BtUnknown = "%.OPAQUE"

btentry BtInt = "@.INTENTRY"
btentry BtFloat = "@.FLOATENTRY"
btentry BtChar = "@.CHARENTRY"
btentry BtInd = "@.INDENTRY"
btentry BtBool = "@.BOOLENTRY"
btentry BtAp = "@.APENTRY"
btentry BtHole = "@.HOLEENTRY"

bttype BtInt = "i32"
bttype BtFloat = "double"
bttype BtChar = "i8"
bttype BtInd = "%.OPAQUE*"
bttype BtBool = "i1"
bttype BtAp = "%.OPAQUE*"
bttype BtHole = "%.OPAQUE*"

ctypeToBt CDouble = BtFloat
ctypeToBt CInt = BtInt
ctypeToBt CChar = BtChar

genNum :: TrM Int
genNum = do
    s <- get
    put (s { snum = 1 + snum s })
    return (snum s)

genId :: TrM String
genId = genNum >>= (return.('i':).show)

genIdL :: String -> TrM String
genIdL pf = genId >>= (return.('%':).(pf++))

getVstack = get >>= (return.vstack)
getVsdelta = get >>= (return.vsdelta)

putVstack st = do
    s <- get
    put (s { vstack = st })

emit :: String -> TrM ()
emit t = do
    s <- get
    put (s { outcode = (t ++ "\n"):outcode s })

initstate = TrState [] 0 [] 0

runTr t = concat $ reverse $ outcode $ execState t initstate

translate prog = runTr (translatep prog)

translatep prog = do
    emit "declare i32 @printf(i8*, ...) nounwind"
    emit "declare i32 @my_readint() nounwind"
    emit "declare double @my_readfloat() nounwind"
    emit "declare i8 @my_readchar() nounwind"
    emit "declare void @my_printint(i32) nounwind"
    emit "declare void @my_printfloat(double) nounwind"
    emit "declare void @my_printchar(i8) nounwind"
    emit "declare double @sin(double) nounwind"
    emit "declare double @cos(double) nounwind"
    emit "declare double @tan(double) nounwind"
    emit "declare double @sinh(double) nounwind"
    emit "declare double @cosh(double) nounwind"
    emit "declare double @tanh(double) nounwind"
    emit "declare double @asin(double) nounwind"
    emit "declare double @acos(double) nounwind"
    emit "declare double @atan(double) nounwind"
    emit "declare double @atan2(double,double) nounwind"
    emit "declare double @ceil(double) nounwind"
    emit "declare double @floor(double) nounwind"
    emit "declare double @round(double) nounwind"
    emit "declare double @abs(double) nounwind"
    emit "declare double @pow(double,double) nounwind"
    emit "declare double @sqrt(double) nounwind"
    emit "declare double @cbrt(double) nounwind"
    emit "declare double @log(double) nounwind"
    emit "declare double @log10(double) nounwind"
    emit "%.ENTRY = type { %.FUNT*, %.FUNT*, %.FUNT2* }"
    emit "%.APNODE = type { %.ENTRY*, %.OPAQUE*, %.OPAQUE* }"
    emit "%.GLOBNODE = type { %.ENTRY*, i32 }"
    emit "%.INTNODE = type { %.ENTRY*, i32 }"
    emit "%.FLOATNODE = type { %.ENTRY*, double }"
    emit "%.CHARNODE = type { %.ENTRY*, i8 }"
    emit "%.INDNODE = type { %.ENTRY*, %.OPAQUE* }"
    emit "%.HOLENODE = type { %.ENTRY*, i32 }"
    emit "%.BOOLNODE = type { %.ENTRY*, i1 }"
    emit "%.OPAQUE = type { %.ENTRY* }"
    emit "%.FUNT = type void (%.OPAQUE**) nounwind"
    emit "%.FUNT2 = type void (%.OPAQUE*) nounwind"
    emit "@.STACKTOP = global %.OPAQUE** zeroinitializer"
    emit "@.APENTRY = constant %.ENTRY {%.FUNT* @.apeval, %.FUNT* @.apunwind, %.FUNT2* @.apprint}"
    emit "@.INTENTRY = constant %.ENTRY {%.FUNT* @.inteval, %.FUNT* @.intunwind, %.FUNT2* @.intprint}"
    emit "@.FLOATENTRY = constant %.ENTRY {%.FUNT* @.inteval, %.FUNT* @.intunwind, %.FUNT2* @.floatprint}"
    emit "@.CHARENTRY = constant %.ENTRY {%.FUNT* @.inteval, %.FUNT* @.intunwind, %.FUNT2* @.charprint}"
    emit "@.BOOLENTRY = constant %.ENTRY {%.FUNT* @.inteval, %.FUNT* @.intunwind, %.FUNT2* @.boolprint}"
    emit "@.INDENTRY = constant %.ENTRY {%.FUNT* @.indeval, %.FUNT* @.indunwind, %.FUNT2* @.indprint}"
    emit "@.HOLEENTRY = constant %.ENTRY {%.FUNT* @.holeeval, %.FUNT* @.holeunwind, %.FUNT2* @.holeprint}"
    emit "declare void @exit(i32)"
    emit "define fastcc void @.holeeval(%.OPAQUE** %STACKDUMP) nounwind {";
    emit "\tcall void @exit(i32 1)";
    emit "\tunreachable"
    emit "}"
    emit "define fastcc void @.holeunwind(%.OPAQUE** %STACKDUMP) nounwind {";
    emit "\tcall void @exit(i32 1)";
    emit "\tunreachable"
    emit "}"
    emit "define fastcc void @.apeval(%.OPAQUE** %STACKDUMP) nounwind {";
    emit "\t%stack = load %.OPAQUE*** @.STACKTOP"
    emit "\ttail call fastcc void @.apunwind(%.OPAQUE** %stack) nounwind"
    emit "\tret void"
    emit "}"
    emit "define fastcc void @.inteval(%.OPAQUE** %STACKDUMP) nounwind {";
    emit "\tret void"
    emit "}"
    emit "define fastcc void @.indeval(%.OPAQUE** %STACKDUMP) nounwind {";
    emit "\t%stack = load %.OPAQUE*** @.STACKTOP"
    emit "\t%topl = load %.OPAQUE** %stack"
    emit "\t%node = bitcast %.OPAQUE* %topl to %.INDNODE*"
    emit "\t%ind = getelementptr %.INDNODE* %node, i32 0, i32 1"
    emit "\t%indl = load %.OPAQUE** %ind"
    emit "\tstore %.OPAQUE* %indl, %.OPAQUE** %stack"
    emit "\t%trgt = bitcast %.OPAQUE* %indl to %.ENTRY**"
    emit "\t%tagl = load %.ENTRY** %trgt"
    emit "\t%fun = getelementptr %.ENTRY* %tagl, i32 0, i32 0"
    emit "\t%funl = load %.FUNT** %fun"
    emit "\ttail call fastcc void %funl(%.OPAQUE** %STACKDUMP) nounwind"
    emit "\tret void"
    emit "}"
    emit "define fastcc void @.indunwind(%.OPAQUE** %STACKDUMP) nounwind {";
    emit "\t%stack = load %.OPAQUE*** @.STACKTOP"
    emit "\t%topl = load %.OPAQUE** %stack"
    emit "\t%node = bitcast %.OPAQUE* %topl to %.INDNODE*"
    emit "\t%ind = getelementptr %.INDNODE* %node, i32 0, i32 1"
    emit "\t%indl = load %.OPAQUE** %ind"
    emit "\tstore %.OPAQUE* %indl, %.OPAQUE** %stack"
    emit "\t%trgt = bitcast %.OPAQUE* %indl to %.ENTRY**"
    emit "\t%tagl = load %.ENTRY** %trgt"
    emit "\t%fun = getelementptr %.ENTRY* %tagl, i32 0, i32 1"
    emit "\t%funl = load %.FUNT** %fun"
    emit "\ttail call fastcc void %funl(%.OPAQUE** %STACKDUMP) nounwind"
    emit "\tret void"
    emit "}"
    emit "define fastcc void @.apunwind(%.OPAQUE** %STACKDUMP) nounwind {";
    emit "\t%stack = load %.OPAQUE*** @.STACKTOP"
    emit "\t%topl = load %.OPAQUE** %stack"
    emit "\t%node = bitcast %.OPAQUE* %topl to %.APNODE*"
    emit "\t%head = getelementptr %.APNODE* %node, i32 0, i32 1"
    emit "\t%headl = load %.OPAQUE** %head"
    emit "\t%stackn = getelementptr %.OPAQUE** %stack, i32 -1"
    emit "\tstore %.OPAQUE* %headl, %.OPAQUE** %stackn"
    emit "\tstore %.OPAQUE** %stackn, %.OPAQUE*** @.STACKTOP"
    emit "\t%tag = bitcast %.OPAQUE* %headl to %.ENTRY**"
    emit "\t%tagl = load %.ENTRY** %tag"
    emit "\t%fun = getelementptr %.ENTRY* %tagl, i32 0, i32 1"
    emit "\t%funl = load %.FUNT** %fun"
    emit "\ttail call fastcc void %funl(%.OPAQUE** %STACKDUMP) nounwind"
    emit "\tret void"
    emit "}"
    emit "define fastcc void @.intunwind(%.OPAQUE** %STACKDUMP) nounwind {";
    emit "\tstore %.OPAQUE** %STACKDUMP, %.OPAQUE*** @.STACKTOP"
    emit "\tret void"
    emit "}"
    emit "@.holestr = internal constant [5 x i8] c\"hole\\00\""
    emit "define fastcc void @.holeprint(%.OPAQUE* %STACKDUMP) nounwind {"
    emit "\tcall i32 (i8*, ...)* @printf(i8* noalias getelementptr ([5 x i8]* @.holestr, i32 0, i32 0))"
    emit "\tret void"
    emit "}"
    emit "@.apstr = internal constant [3 x i8] c\"ap\\00\""
    emit "define fastcc void @.apprint(%.OPAQUE* %node) nounwind {"
    emit "\t%apnode = bitcast %.OPAQUE* %node to %.APNODE*"
    emit "\t%leftv = getelementptr %.APNODE* %apnode, i32 0, i32 1"
    emit "\t%lleftv = load %.OPAQUE** %leftv"
    emit "\t%entryleft = getelementptr %.OPAQUE* %lleftv, i32 0, i32 0"
    emit "\t%lentryleft = load %.ENTRY** %entryleft"
    emit "\t%printleft = getelementptr %.ENTRY* %lentryleft, i32 0, i32 2"
    emit "\t%lprintleft = load %.FUNT2** %printleft"
    emit "\tcall fastcc void %lprintleft(%.OPAQUE* %lleftv) nounwind"
    emit "\t%rightv = getelementptr %.APNODE* %apnode, i32 0, i32 2"
    emit "\t%lrightv = load %.OPAQUE** %rightv"
    emit "\t%entryright = getelementptr %.OPAQUE* %lrightv, i32 0, i32 0"
    emit "\t%lentryright = load %.ENTRY** %entryright"
    emit "\t%printright = getelementptr %.ENTRY* %lentryright, i32 0, i32 2"
    emit "\t%lprintright = load %.FUNT2** %printright"
    emit "\tcall fastcc void %lprintright(%.OPAQUE* %lrightv) nounwind"
    emit "\tcall i32 (i8*, ...)* @printf(i8* noalias getelementptr ([3 x i8]* @.apstr, i32 0, i32 0))"
    emit "\tret void"
    emit "}"
    emit "@.intstr = internal constant [6 x i8] c\"int%d\\00\""
    emit "define fastcc void @.intprint(%.OPAQUE* %node) nounwind {"
    emit "\t%intnode = bitcast %.OPAQUE* %node to %.INTNODE*"
    emit "\t%intv = getelementptr %.INTNODE* %intnode, i32 0, i32 1"
    emit "\t%lintv = load i32* %intv"
    emit "\tcall i32 (i8*, ...)* @printf(i8* noalias getelementptr ([6 x i8]* @.intstr, i32 0, i32 0), i32 %lintv)"
    emit "\tret void"
    emit "}"
    emit "@.charstr = internal constant [7 x i8] c\"char%c\\00\""
    emit "define fastcc void @.charprint(%.OPAQUE* %node) nounwind {"
    emit "\t%charnode = bitcast %.OPAQUE* %node to %.CHARNODE*"
    emit "\t%charv = getelementptr %.CHARNODE* %charnode, i32 0, i32 1"
    emit "\t%lcharv = load i8* %charv"
    emit "\tcall i32 (i8*, ...)* @printf(i8* noalias getelementptr ([7 x i8]* @.charstr, i32 0, i32 0), i8 %lcharv)"
    emit "\tret void"
    emit "}"
    emit "define fastcc void @.indprint(%.OPAQUE* %node) nounwind {"
    emit "\t%indnode = bitcast %.OPAQUE* %node to %.INDNODE*"
    emit "\t%leftv = getelementptr %.INDNODE* %indnode, i32 0, i32 1"
    emit "\t%lleftv = load %.OPAQUE** %leftv"
    emit "\t%entryleft = getelementptr %.OPAQUE* %lleftv, i32 0, i32 0"
    emit "\t%lentryleft = load %.ENTRY** %entryleft"
    emit "\t%printleft = getelementptr %.ENTRY* %lentryleft, i32 0, i32 2"
    emit "\t%lprintleft = load %.FUNT2** %printleft"
    emit "\tcall fastcc void %lprintleft(%.OPAQUE* %lleftv) nounwind"
    emit "\tret void"
    emit "}"
    emit "define fastcc void @.floatprint(%.OPAQUE* %STACKDUMP) nounwind {"
    emit "\tret void"
    emit "}"
    emit "define fastcc void @.boolprint(%.OPAQUE* %STACKDUMP) nounwind {"
    emit "\tret void"
    emit "}"
    mapM_ translate' prog
    emit "define i32 @prog(%.OPAQUE** %STACKDUMP) nounwind {"
    emit "\tstore %.OPAQUE** %STACKDUMP, %.OPAQUE*** @.STACKTOP"
    translatec (PushGlobal "prog")
    translatec (Eval){-
    emit "\t%stack = load %.OPAQUE*** @.STACKTOP"
    emit "\t%ltop = load %.OPAQUE** %stack"
    emit "\t%entry = getelementptr %.OPAQUE* %ltop, i32 0, i32 0"
    emit "\t%lentry = load %.ENTRY** %entry"
    emit "\t%print = getelementptr %.ENTRY* %lentry, i32 0, i32 2"
    emit "\t%lprint = load %.FUNT2** %print"
    emit "\tcall fastcc void %lprint(%.OPAQUE* %ltop) nounwind"-}
    emit "\tret i32 0"
    emit "}"

translate' (GFun id n c) = do
    emit ("@e" ++ id ++ " = constant %.ENTRY {%.FUNT* @eval" ++ id ++ ", %.FUNT* @unwind" ++ id ++ ", %.FUNT2* @print" ++ id ++ "}")
    emit ("define fastcc void @fun" ++ id ++ "(%.OPAQUE** %STACKDUMP) nounwind {")
    mapM_ translatec c
    emit "}"
    emit ("define fastcc void @eval" ++ id ++ "(%.OPAQUE** %STACKDUMP) nounwind {")
    when (n == 0) $ do
        emit "\t%stack = load %.OPAQUE*** @.STACKTOP"
        emit ("\ttail call fastcc void @fun" ++ id ++ "(%.OPAQUE** %stack) nounwind")
    emit "\tret void"
    emit "}"
    emit ("define fastcc void @unwind" ++ id ++ "(%.OPAQUE** %STACKDUMP) nounwind {")
    emit "\t%stack = load %.OPAQUE*** @.STACKTOP"
    emit ("\t%below = getelementptr %.OPAQUE** %stack, i32 " ++ show n)
    emit "\t%toofew = icmp ugt %.OPAQUE** %below, %STACKDUMP"
    emit "\tbr i1 %toofew, label %tf, label %ok"
    emit "tf:"
    emit "\tstore %.OPAQUE** %STACKDUMP, %.OPAQUE*** @.STACKTOP"
    emit "\tret void"
    emit "ok:"
    forM_ [1..n] $ \k -> do
        let i = show k
        emit ("\t%pos" ++ i ++ " = getelementptr %.OPAQUE** %stack, i32 " ++ show k)
        emit ("\t%posl" ++ i ++ " = load %.OPAQUE** %pos" ++ i)
        emit ("\t%appn" ++ i ++ " = bitcast %.OPAQUE* %posl" ++ i ++ " to %.APNODE*")
        emit ("\t%apparg" ++ i ++ " = getelementptr %.APNODE* %appn" ++ i ++ ", i32 0, i32 2")
        emit ("\t%appargl" ++ i ++ " = load %.OPAQUE** %apparg" ++ i)
        emit ("\t%poss" ++ i ++ " = getelementptr %.OPAQUE** %stack, i32 " ++ show (k-1))
        emit ("\tstore %.OPAQUE* %appargl" ++ i ++ ", %.OPAQUE** %poss" ++ i)
    emit ("\ttail call fastcc void @fun" ++ id ++ "(%.OPAQUE** %STACKDUMP) nounwind")
    emit "\tret void"
    emit "}"
    emit ("define fastcc void @print" ++ id ++ "(%.OPAQUE* %STACKDUMP) nounwind {")
    emit ("\tret void")
    emit "}"
    emit ("@node" ++ id ++ " = global %.GLOBNODE {%.ENTRY* @e" ++ id ++", i32 0}")

translatec (Push k) = do
    val <- load_from_stack BtUnknown k
    movestackptr 1
    store_on_stack BtUnknown 0 val
translatec (Pop k) = movestackptr (-k)
translatec (Slide k) = do
    val <- load_from_stack BtUnknown 0
    store_on_stack BtUnknown k val
    movestackptr (-k)
translatec (Alloc k) = do
    movestackptr k
    forM_ [0..k-1] $ \l -> do
        n <- buildnode BtHole
        store_on_stack BtHole l n
translatec (PushInt v) = do
    n <- buildnode BtInt
    set_node_field1 BtInt n (show v)
    movestackptr 1
    store_on_stack BtInt 0 n
translatec (PushFloat v) = do
    n <- buildnode BtFloat
    set_node_field1 BtFloat n (show v)
    movestackptr 1
    store_on_stack BtFloat 0 n
translatec (PushBool v) = do
    n <- buildnode BtBool
    set_node_field1 BtBool n (if v then "1" else "0")
    movestackptr 1
    store_on_stack BtBool 0 n
translatec (PushChar v) = do
    n <- buildnode BtChar
    set_node_field1 BtChar n (show $ max 0 $ min 255 $ ord v)
    movestackptr 1
    store_on_stack BtChar 0 n
translatec (PushGlobal n) = do
    movestackptr 1
    node <- llvm_bitcast "%.GLOBNODE*" ("@node" ++ n) "%.OPAQUE*"
    store_on_stack BtUnknown 0 node
translatec (Update k) = do
    n <- load_from_stack BtInd k
    set_node_tag BtInd n (btentry BtInd)
    val <- load_from_stack BtUnknown 0
    set_node_field1 BtInd n val 
    store_on_stack BtUnknown k val
    movestackptr (-1)
translatec (MkAp) = do
    v <- load_from_stack BtUnknown 0
    w <- load_from_stack BtUnknown 1
    n <- buildnode BtAp
    set_node_field1 BtAp n v
    set_node_field2 BtAp n w
    store_on_stack BtAp 1 n
    movestackptr (-1)
translatec Eval = do
    n <- load_from_stack BtUnknown 0
    tag <- get_node_tag BtUnknown n
    fun <- llvm_getelementptr "%.ENTRY*" tag ["0","0"]
    funl <- llvm_load "%.FUNT*" fun
    flush_vstack
    emit ("\tcall fastcc void " ++ funl ++ "(%.OPAQUE** %STACKDUMP) nounwind")
translatec Unwind = do
    n <- load_from_stack BtUnknown 0
    tag <- get_node_tag BtUnknown n
    fun <- llvm_getelementptr "%.ENTRY*" tag ["0","1"]
    funl <- llvm_load "%.FUNT*" fun
    flush_vstack
    emit ("\tcall fastcc void " ++ funl ++ "(%.OPAQUE** %STACKDUMP) nounwind")
    emit ("\tret void")
translatec Return = do
    flush_vstack
    emit ("\tret void")
translatec (Label l) = emit (l ++ ":")
translatec (Jump l) = do
    flush_vstack
    emit ("\tbr label %" ++ l)
translatec (JFalse l) = do
    i <- genId
    n <- load_from_stack BtBool 0
    cond <- get_node_field1 BtBool n
    movestackptr (-1)
    flush_vstack
    emit ("\tbr i1 " ++ cond ++ ", label %" ++ i ++ ", label %" ++ l)
    emit (i ++ ":")
translatec (CallCFun mrt ats nm) = do
    let rtn = case mrt of
                Just rt -> bttype (ctypeToBt rt)
                Nothing -> "void"
    let bats = map ctypeToBt ats
    argns <- mapM (uncurry load_from_stack) (zip bats [0..])
    argvs <- mapM (uncurry get_node_field1) (zip bats argns)
    res <- llvm_call rtn (map bttype bats) nm argvs
    movestackptr (-length ats)
    case mrt of
        Just rt -> do
            let brt  = ctypeToBt rt
            n <- buildnode brt
            set_node_field1 brt n res
            movestackptr 1
            store_on_stack brt 0 n
        Nothing -> return ()
translatec AddInt = translate_builtin_bin BtInt BtInt (llvm_bin "add" "i32")
translatec SubInt = translate_builtin_bin BtInt BtInt (llvm_bin "sub" "i32")
translatec MulInt = translate_builtin_bin BtInt BtInt (llvm_bin "mul" "i32")
translatec DivInt = translate_builtin_bin BtInt BtInt (llvm_bin "sdiv" "i32")
translatec RemInt = translate_builtin_bin BtInt BtInt (llvm_bin "srem" "i32")
translatec NegInt = translate_builtin_un BtInt BtInt (llvm_bin "sub" "i32" "0")
translatec EqInt = 
    translate_builtin_bin BtInt BtBool (llvm_bin "icmp eq" "i32")
translatec NEInt = 
    translate_builtin_bin BtInt BtBool (llvm_bin "icmp ne" "i32")
translatec LTInt = 
    translate_builtin_bin BtInt BtBool (llvm_bin "icmp slt" "i32")
translatec LEInt = 
    translate_builtin_bin BtInt BtBool (llvm_bin "icmp sle" "i32") 
translatec AddFloat = 
    translate_builtin_bin BtFloat BtFloat (llvm_bin "add" "double")
translatec SubFloat = 
    translate_builtin_bin BtFloat BtFloat (llvm_bin "sub" "double")
translatec MulFloat = 
    translate_builtin_bin BtFloat BtFloat (llvm_bin "mul" "double")
translatec DivFloat = 
    translate_builtin_bin BtFloat BtFloat (llvm_bin "fdiv" "double")
translatec RemFloat = 
    translate_builtin_bin BtFloat BtFloat (llvm_bin "frem" "double")
translatec NegFloat = 
    translate_builtin_un BtFloat BtFloat (llvm_bin "sub" "double" "0.0")
translatec EqFloat = 
    translate_builtin_bin BtFloat BtBool (llvm_bin "fcmp oeq" "double")
translatec NEFloat = 
    translate_builtin_bin BtFloat BtBool (llvm_bin "fcmp one" "double")
translatec LTFloat = 
    translate_builtin_bin BtFloat BtBool (llvm_bin "fcmp olt" "double")
translatec LEFloat = 
    translate_builtin_bin BtFloat BtBool (llvm_bin "fcmp ole" "double")
translatec AndBool = translate_builtin_bin BtBool BtBool (llvm_bin "and" "i1")
translatec OrBool = translate_builtin_bin BtBool BtBool (llvm_bin "or" "i1")
translatec NotBool = 
    translate_builtin_un BtBool BtBool (llvm_bin "xor" "i1" "1")
translatec EqvBool = 
    translate_builtin_bin BtBool BtBool (llvm_bin "icmp eq" "i1")
translatec NEvBool = 
    translate_builtin_bin BtBool BtBool (llvm_bin "icmp ne" "i1")
translatec EqChar = 
    translate_builtin_bin BtChar BtBool (llvm_bin "icmp eq" "i8")
translatec NEChar = 
    translate_builtin_bin BtChar BtBool (llvm_bin "icmp ne" "i8")
translatec LTChar = 
    translate_builtin_bin BtChar BtBool (llvm_bin "icmp ult" "i8")
translatec LEChar = 
    translate_builtin_bin BtChar BtBool (llvm_bin "icmp ule" "i8") 
translatec IntToFloat =
    translate_builtin_un BtInt BtFloat (flip (llvm_sitofp "i32") "double")
translatec FloatToInt =
    translate_builtin_un BtFloat BtInt (flip (llvm_fptosi "double") "i32")

llvm_getelementptr ty arg idx = do
    i <- genIdL "geptr"
    emit ("\t" ++ i ++ " = getelementptr " ++ ty ++ " " ++ arg ++ ", " ++ concat (intersperse ", " ["i32 " ++ k | k <- idx]))
    return i

llvm_malloc ty = do
    i <- genIdL "mall"
    emit ("\t" ++ i ++ " = malloc " ++ ty)
    return i

llvm_store ty src trgt = do
    emit ("\tstore " ++ ty ++ " " ++ src ++ ", " ++ ty ++ "* " ++ trgt)

llvm_load ty src = do
    i <- genIdL "load"
    emit ("\t" ++ i ++ " = load " ++ ty ++ "* " ++ src)
    return i

llvm_bin op ty a1 a2 = do
    i <- genIdL "bin"
    emit ("\t" ++ i ++ " = " ++ op ++ " " ++ ty ++ " " ++ a1 ++ ", " ++ a2)
    return i

llvm_bitcast ty arg ty' | ty == ty' = return arg
                        | otherwise = do
    i <- genIdL "bcast"
    emit ("\t" ++ i ++ " = bitcast " ++ ty ++ " " ++ arg ++ " to " ++ ty')
    return i
    
llvm_call rt ats nm args = do
    i <- genIdL "call"
    emit ((if rt == "void" then "\t" else ("\t" ++ i ++ " = ")) ++ "call " ++ rt ++ "(" ++ concat (intersperse ", " ats) ++ ")* @" ++ nm ++ "(" ++ concat (intersperse ", " [t ++ " " ++ a | (t,a) <- zip ats args]) ++ ")")
    return i

llvm_sitofp st arg tt = do
    i <- genIdL "sitofp"
    emit ("\t" ++ i ++ " = sitofp " ++ st ++ " " ++ arg ++ " to " ++ tt)
    return i

llvm_fptosi st arg tt = do
    i <- genIdL "fptosi"
    emit ("\t" ++ i ++ " = fptosi " ++ st ++ " " ++ arg ++ " to " ++ tt)
    return i

buildnode t = do
    i <- llvm_malloc (btnode t)
    set_node_tag t i (btentry t)
    return i

get_node_tag ty arg = do
    trgt <- llvm_getelementptr (btnode ty ++ "*") arg ["0","0"]
    llvm_load ("%.ENTRY*") trgt

get_node_field k ty arg = do
    trgt <- llvm_getelementptr (btnode ty ++ "*") arg ["0",show k]
    llvm_load (bttype ty) trgt

get_node_field1 = get_node_field 1
get_node_field2 = get_node_field 2

set_node_tag ty arg tag = do
    trgt <- llvm_getelementptr (btnode ty ++ "*") arg ["0","0"]
    llvm_store ("%.ENTRY*") tag trgt

set_node_field k ty arg val = do
    i <- genId
    trgt <- llvm_getelementptr (btnode ty ++ "*") arg ["0",show k]
    llvm_store (bttype ty) val trgt

set_node_field1 = set_node_field 1
set_node_field2 = set_node_field 2

load_from_vstack k = do
    s <- getVstack
    return (kth s k) where
        kth [] _ = Nothing
        kth (h:_) 0 = h
        kth (_:t) k = kth t (k-1)

load_from_stack ty k = do
    vd <- load_from_vstack k
    val <- case vd of
        Just v -> return v
        Nothing -> do
            i <- genIdL "lfs"
            stack <- llvm_load "%.OPAQUE**" "@.STACKTOP"
            vsd <- getVsdelta
            ptr <- llvm_getelementptr "%.OPAQUE**" stack [show (k-vsd)]
            llvm_load "%.OPAQUE*" ptr
    llvm_bitcast "%.OPAQUE*" val (btnode ty ++ "*")

store_on_stack ty k val = do
    opval <- llvm_bitcast (btnode ty ++ "*") val "%.OPAQUE*"
    s <- getVstack
    putVstack (kth s k opval) where
        kth [] 0 v = [Just v]
        kth [] k v = Nothing:kth [] (k-1) v
        kth (_:t) 0 v = Just v:t
        kth (h:t) k v = h:kth t (k-1) v

movestackptr k = do
    s <- get
    put (s { vsdelta = vsdelta s + k, vstack = trf (vstack s) }) where
        trf st = if k <= 0 then drop (-k) st else replicate k Nothing ++ st

flush_vstack = do
    s <- get
    put (s { vsdelta = 0, vstack = [] })
    stack <- llvm_load "%.OPAQUE**" "@.STACKTOP"
    moved <- llvm_getelementptr "%.OPAQUE**" stack [show (-(vsdelta s))]
    llvm_store "%.OPAQUE**" moved "@.STACKTOP"
    trf (vstack s) moved 0 where
        trf [] _ _ = return ()
        trf (Nothing:t) top k = trf t top (k+1)
        trf (Just v:t) top k = do
            ptr <- llvm_getelementptr "%.OPAQUE**" top [show k]
            llvm_store "%.OPAQUE*" v ptr
            trf t top (k+1)

translate_builtin_un tin tout genseq = do
    argna <- load_from_stack tin 0
    argva <- get_node_field1 tin argna
    res <- genseq argva
    n <- buildnode tout
    set_node_field1 tout n res
    store_on_stack tout 0 n

translate_builtin_bin tin tout genseq = do
    argna <- load_from_stack tin 0
    argnb <- load_from_stack tin 1
    argva <- get_node_field1 tin argna
    argvb <- get_node_field1 tin argnb
    res <- genseq argva argvb
    n <- buildnode tout
    set_node_field1 tout n res
    store_on_stack tout 1 n
    movestackptr (-1)

