module Language(Id(..), Term(..), freevars,
    subst, ssubst, substv, bound, unbound, boundAll, boundSk, unboundSk,
    local, global, builtin) where

import Data.Set(Set)
import qualified Data.Set as Set
import Data.Map(Map)
import qualified Data.Map as Map
import Control.Monad
import Control.Monad.Error
import Control.Monad.State

data Id = Local Int | Global String | Builtin String | Skolem Int deriving (Show, Eq, Ord)

data Term = App Term Term
          | AppTy Term Term
          | Lam Term
          | Pi Term Term
          | Let Term Term
          | Ann Term Term
          | Star
          | Free Id
          | Bound Int
          | IntV Int
          | CharV Char
          | BoolV Bool
          | FloatV Double
          | CaseType Term Term Term Term Term Term Term Term Term Term Term
          | PrimLam String Term
          | PrimPi String Term Term
          | PrimLet [(String,Term)] Term
          deriving (Show, Eq, Ord)

infixr 1 `Pi`
infixl 1 `App`
infixl 1 `AppTy`

newtype DSub = DSub { undsub :: Term -> Term }

instance Show DSub where
    show _ = "dsub"

instance Eq DSub where
    _ == _ = True

instance Ord DSub where
    _ <= _ = True

ssubst :: (Int -> Term -> Term) -> Int -> Term -> Term
ssubst f i (App t1 t2) = App (ssubst f i t1) (ssubst f i t2)
ssubst f i (AppTy t1 t2) = AppTy (ssubst f i t1) (ssubst f i t2)
ssubst f i (Lam t) = Lam (ssubst f (i+1) t)
ssubst f i (Pi t1 t2) = Pi (ssubst f i t1) (ssubst f (i+1) t2)
ssubst f i (Let t1 t2) = Let (ssubst f i t1) (ssubst f (i+1) t2)
ssubst f i (Ann t1 t2) = Ann (ssubst f i t1) (ssubst f i t2)
ssubst f i (CaseType v t1 t2 t3 t4 t5 t6 t7 t8 t9 t10) = 
    CaseType (ssubst f i v) (ssubst f i t1) (ssubst f i t2) (ssubst f i t3)
        (ssubst f i t4) (ssubst f i t5) (ssubst f i t6) (ssubst f i t7)
        (ssubst f i t8) (ssubst f i t9) (ssubst f i t10)
ssubst f i t = f i t

subst :: Int -> Term -> Term -> Term
subst i e = ssubst f i where
    f i (Bound j) | i == j = e
    f i t = t

unbound :: Int -> Term -> Term
unbound d = subst 0 (Free (Local d))

bound :: Int -> Term -> Term
bound d = ssubst f 0 where
    f i (Free (Local d')) | d == d' = Bound i
    f i t = t

boundAll :: Int -> Term -> Term
boundAll = ssubst f where
    f i (Free (Local d)) | i > d = Bound (i - d - 1)
    f i t = t

unboundSk :: Int -> Term -> Term
unboundSk d = subst 0 (Free (Skolem d))

boundSk :: Int -> Term -> Term
boundSk d = ssubst f 0 where
    f i (Free (Skolem d')) | d == d' = Bound i
    f i t = t

substv :: Id -> Term -> Term -> Term
substv n e = ssubst f 0 where
    f i (Free n') | n == n' = e
    f i t = t

homoExtSet :: Ord a => (Term -> Set a) -> Term -> Set a
homoExtSet f (App t1 t2) = homoExtSet f t1 `Set.union` homoExtSet f t2
homoExtSet f (AppTy t1 t2) = homoExtSet f t1 `Set.union` homoExtSet f t2
homoExtSet f (Lam t) = homoExtSet f t
homoExtSet f (Pi t1 t2) = homoExtSet f t1 `Set.union` homoExtSet f t2
homoExtSet f (Let t1 t2) = homoExtSet f t1 `Set.union` homoExtSet f t2
homoExtSet f (Ann t1 t2) = homoExtSet f t1 `Set.union` homoExtSet f t2
homoExtSet f (CaseType v t1 t2 t3 t4 t5 t6 t7 t8 t9 t10) =
    foldl1 (Set.union) (map (homoExtSet f) [v,t1,t2,t3,t4,t5,t6,t7,t8,t9,t10])
homoExtSet f (PrimLam _ t) = homoExtSet f t
homoExtSet f (PrimPi _ t1 t2) = homoExtSet f t1 `Set.union` homoExtSet f t2
homoExtSet f (PrimLet d t) = 
    foldr Set.union (homoExtSet f t) (map (homoExtSet f . snd) d)
homoExtSet f t = f t

freevars :: Term -> Set Id
freevars = homoExtSet f where
    f (Free i) = Set.singleton i
    f _ = Set.empty

builtin = Free . Builtin
local = Free . Local
global = Free . Global

