module Lexer(Token(..), tokenize) where

import ArrowParser
import Control.Arrow
import qualified Data.Set as Set

-- Tokenizer

keywords = Set.fromList ["fn","if","then","else","let","in","casetype","of"]
layoutKwd = Set.fromList ["let"]

data Token = TokInt (Attr Int)
           | TokFloat (Attr Rational) 
           | TokString (Attr String)
           | TokChar (Attr Char)
           | TokName (Attr String)
           | TokPhold (Attr String)
           | TokKwd String
           | TokOp String -- Attr String, jak bedzie mozna definiowac op-y
           | TokLParen | TokRParen 
           | TokLBrace | TokRBrace 
           | TokLBracket | TokRBracket
           | TokComma | TokSemicolon
           | TokBacktick
           deriving (Show, Eq, Ord)

data TokSt = TokSt { row :: Int, col :: Int } deriving Show

decimal = many1 digit
octal = many1 (oneOf ['0'..'7'])
hexadecimal = many1 (digit <|> oneOf ['a'..'f'] <|> oneOf ['A'..'F'])

number = (skip.tokenP) '0' >>> 
           ((skip.oneOf) ['x','X'] >>> hexint 
        <|> (skip.oneOf) ['o','O'] >>> octint
        <|> arr (const '0') >>> numberRest)
     <|> oneOf "123456789" >>> numberRest

hexint = hexadecimal >>^ (TokInt . Attr . read . ("0x"++))
octint = octal       >>^ (TokInt . Attr . read . ("0o"++))

numberRest = (emptyP &&& (ignore >>> many digit) >>^ (read . uncurry (:) :: (Char,String) -> Int)) >>>
        (emptyP &&& (ignore >>> (skip.tokenP) '.' >>> decimal)
            >>^ (\(a,b) -> TokFloat (Attr (fromIntegral a+mkrat b)))
     <|> arr (TokInt . Attr))
     where mkrat x = fromIntegral(read x) / 10^length x

exponent = (skip.oneOf) ['e','E']

labelKw w | w `Set.member` keywords = TokKwd w
          | otherwise = TokName (Attr w)

ident = letter `thenMany` alphaNum 

name = ident >>^ labelKw

placeholder = tokenP '_' &&& ident >>^ (TokPhold . Attr . uncurry (:))

op = (many1 symbol >>^ (TokOp))

symbol = oneOf ":!#$%&*+./<>=?@^|-~\\"

special = oneOf (map fst specials)

graphic = alphaNum <|> symbol <|> special

inchar = escape <|> graphic <|> oneOf "\" "

instr = escape <|> graphic <|> oneOf "' "

escape = (iskip.tokenP) '\\' >>> 
            (tokenP '\\' 
         <|> tokenP 'n' >>^ const '\n')

char = between (tokenP '\'') (tokenP '\'') inchar >>^ (TokChar . Attr) 

str = between (tokenP '"') (tokenP '"') (many1 instr) >>^ (TokString . Attr) 

specialt (c, t) = tokenP c >>^ const t

specials = [('(',TokLParen),(')',TokRParen),
            ('[',TokLBracket),(']',TokRBracket),
            ('{',TokLBrace),('}',TokRBrace),
            (',',TokComma),(';',TokSemicolon),
            ('`',TokBacktick)]

tok = number <|> name <|> placeholder <|> op <|> char <|> str
  <|> choice (map specialt specials)

ctok = proc () -> do
    p1 <- currPos -< ()
    t <- tok -< ()
    p2 <- currPos -< ()
    st <- getState -< ()
    putState -< st { col = col st + (p2 - p1) }
    returnA -< (t, col st, row st)

wspacecol ts c [] = c
wspacecol ts c (' ':t) = wspacecol ts (c+1) t
wspacecol ts c ('\t':t) = wspacecol ts (c+ts-1 - (c-1) `mod` 4) t

skipwspace = proc a -> do
    wl <- many (oneOf " \t") -< ()
    st <- getState -< ()
    putState -< st { col = wspacecol 4 (col st) wl }
    returnA -< a

newrow = proc a -> do
    st <- getState -< ()
    putState -< st { col = 0, row = row st + 1 }
    returnA -< a

progline = skipwspace >>> (linetail <|> consA ctok >>> progline <|> emptyP)
linetail = (iskip.tokenP) '#' >>> skipMany (ignore >>> noneOf "\r\n")

tokenizer = progline >>> 
    ((iskip.tokenP) '\n' >>> newrow >>> tokenizer <|> emptyP)

tokenize s = case parseAll tokenizer ([],s,TokSt 0 0) of
    Left x -> Left x
    Right (a,b) -> Right (reverse a)

