module Parser(parseExpr) where

import ArrowParser
import Lexer
import Language
import Control.Arrow
import Data.Set(Set)
import qualified Data.Set as Set

data PState = PState {
    snum :: Int,
    sname :: String
}
type PArrL = PArr PState Token

initstate sn = PState 0 sn

builtins = Set.fromList [
    "addint","subint","mulint","divint","remint","negint",
    "eqint", "neint", "ltint", "leint",
    "addfloat","subfloat","mulfloat","divfloat","remfloat","negfloat",
    "eqfloat", "nefloat", "ltfloat", "lefloat",
    "inttofloat", "floattoint",
    "sin","cos","tan","asin","acos","atan","atan2",
    "sinh","cosh","tanh",
    "pow","sqrt","cbrt","log","log10",
    "ceil","floor","round","abs",
    "and","or","not","eqv","neqv","if",
    "unit","true","false",
    "inl","inr","case",
    "fst","snd","pair",
    "Arrow","Pi","Int","Char","Float","Bool","Pair","Sum","Star","Unit",
    "fix","Fix","wrap","unwrap",
    "return","bind","readint","readchar","readfloat","IO",
    "printint","printchar","printfloat"]

genNum :: PArrL () Int
genNum = proc () -> do
    s <- getState -< ()
    putState -< s { snum = snum s + 1 }
    returnA -< snum s

genName :: PArrL () String
genName = proc () -> do
    s <- getState -< ()
    putState -< s { snum = snum s + 1 }
    returnA -< show (snum s) ++ "_" ++ sname s

ifExpr = proc x -> do
    tokenP (TokKwd "if") -< ()
    cond <- expr0 -< ()
    tokenP (TokKwd "then") -< ()
    e1 <- expr0 -< ()
    tokenP (TokKwd "else") -< ()
    e2 <- expr0 -< ()
    returnA -< (builtin "if" `App` cond `AppTy` e1 `App` e2)

fnExpr = proc x -> do
    tokenP (TokKwd "fn") -< ()
    (TokName (Attr n)) <- tokenP (TokName AttrPH) -< ()
    tokenP (TokOp "=>") -< ()
    e <- expr -< ()
    returnA -< (PrimLam n e)

letExpr = proc x -> do
    tokenP (TokKwd "let") -< ()
    d <- defs -< ()
    tokenP (TokKwd "in") -< ()
    e2 <- expr -< ()
    returnA -< (PrimLet d e2)

piExpr = proc x -> do
    tokenP (TokLBrace) -< ()
    (TokName (Attr n)) <- tokenP (TokName AttrPH) -< ()
    e1 <- option Star ((skip.tokenP) (TokOp "::") >>> expr) -< ()
    tokenP (TokRBrace) -< ()
    e2 <- expr -< ()
    returnA -< (PrimPi n e1 e2)
{-
listExpr = proc x -> do
    tokenP (TokLBracket) -< ()
    l <- sepBy expr ((skip.tokenP) TokComma) -< ()
    tokenP (TokRBracket) -< ()
    mv <- newMetavar -< ()
    returnA -< buildList mv l

buildList mv [] = global "nil" `App` mv
buildList mv (h:t) = global "cons" `App` mv `App` h `App` buildList mv t
-}
whichvar k | k `Set.member` builtins = Builtin k
           | otherwise = Global k

casetypeExpr = proc x -> do
    tokenP (TokKwd "casetype") -< ()
    (TokName (Attr n)) <- tokenP (TokName AttrPH)  -< ()
    tokenP (TokKwd "of") -< ()
    tokenP (TokLBrace) -< ()
    l <- sepBy1 casetypeOpt ((skip.tokenP) TokSemicolon) -< ()
    tokenP (TokRBrace) -< ()
    let mv = [global "undefined" `App` m | m <- [Star,Star,Star,Star,Star,Star]] ++
             [PrimLam "x" (PrimLam "y" (global "undefined" `App` m)) | 
                m <- [Star,Star,Star,Star]]
    returnA -< casetypeGen (Free (whichvar n)) (map Left mv) l

casetypeOpt = (expr >>> (iskip.tokenP) (TokOp "=>")) &&& expr

casetypeGen n = ctGen where
    ct = either id id
    ctGen [a1,a2,a3,a4,a5,a6,a7,a8,a9,a10] [] =
        CaseType n (ct a1) (ct a2) (ct a3) (ct a4) (ct a5) (ct a6)
            (ct a7) (ct a8) (ct a9) (ct a10)
    ctGen [Left _,a2,a3,a4,a5,a6,a7,a8,a9,a10] ((Free (Builtin "Star"),e):t) =
        ctGen [Right e,a2,a3,a4,a5,a6,a7,a8,a9,a10] t
    ctGen [a1,Left _,a3,a4,a5,a6,a7,a8,a9,a10] ((Free (Builtin "Unit"),e):t) =
        ctGen [a1,Right e,a3,a4,a5,a6,a7,a8,a9,a10] t
    ctGen [a1,a2,Left _,a4,a5,a6,a7,a8,a9,a10] ((Free (Builtin "Bool"),e):t) =
        ctGen [a1,a2,Right e,a4,a5,a6,a7,a8,a9,a10] t
    ctGen [a1,a2,a3,Left _,a5,a6,a7,a8,a9,a10] ((Free (Builtin "Int"),e):t) =
        ctGen [a1,a2,a3,Right e,a5,a6,a7,a8,a9,a10] t
    ctGen [a1,a2,a3,a4,Left _,a6,a7,a8,a9,a10] ((Free (Builtin "Float"),e):t) =
        ctGen [a1,a2,a3,a4,Right e,a6,a7,a8,a9,a10] t
    ctGen [a1,a2,a3,a4,a5,Left _,a7,a8,a9,a10] ((Free (Builtin "Char"),e):t) =
        ctGen [a1,a2,a3,a4,a5,Right e,a7,a8,a9,a10] t
    ctGen [a1,a2,a3,a4,a5,a6,Left _,a8,a9,a10] 
        ((App (App (Free (Builtin "Pair")) 
            (Free (Global v1))) (Free (Global v2)),e):t) =
        ctGen [a1,a2,a3,a4,a5,a6,Right (PrimLam v1 (PrimLam v2 e)),a8,a9,a10] t
    ctGen [a1,a2,a3,a4,a5,a6,a7,Left _,a9,a10] 
        ((App (App (Free (Builtin "Sum")) 
            (Free (Global v1))) (Free (Global v2)),e):t) =
        ctGen [a1,a2,a3,a4,a5,a6,a7,Right (PrimLam v1 (PrimLam v2 e)),a9,a10] t
    ctGen [a1,a2,a3,a4,a5,a6,a7,a8,Left _,a10] 
        ((App (App (Free (Builtin "Arrow")) 
            (Free (Global v1))) (Free (Global v2)),e):t) =
        ctGen [a1,a2,a3,a4,a5,a6,a7,a8,Right (PrimLam v1 (PrimLam v2 e)),a10] t
    ctGen [a1,a2,a3,a4,a5,a6,a7,a8,a9,Left _] 
        ((PrimPi n (Free (Global v1)) (Free (Global v2)),e):t) =
        ctGen [a1,a2,a3,a4,a5,a6,a7,a8,a9,Right (PrimLam v1 (PrimLam v2 e))] t

atom :: PArrL () Term
atom = tokenP (TokInt AttrPH) >>^ (\(TokInt (Attr i)) -> IntV i)
   <|> tokenP (TokFloat AttrPH) >>^ (\(TokFloat (Attr i)) -> FloatV (fromRational i))
   <|> tokenP (TokName AttrPH) >>^ (\(TokName (Attr n)) -> Free (whichvar n))
   <|> tokenP (TokChar AttrPH) >>^ (\(TokChar (Attr c)) -> CharV c)
   <|> tokenP (TokString AttrPH) >>^ (\(TokString (Attr c)) -> foldr (\x y -> global "cons" `App` builtin "Char" `App` CharV x `App` y) (global "nil" `App` builtin "Char") c)
--   <|> tokenP (TokPhold AttrPH) >>> ignore >>> newMetavar
   <|> between (tokenP TokLParen) (tokenP TokRParen) (expr <|> arr (const $ builtin "unit"))
   <|> ifExpr
   <|> fnExpr
   <|> letExpr
   <|> piExpr
   <|> casetypeExpr

expr9 = atom

expr8 = expr9 >>> chain (emptyP &&& (const () ^>> expr9) >>^ (uncurry App))

expr7 = (skip.tokenP) (TokOp "-") >>> expr7 >>> unexpr "neg"
    <|> (skip.tokenP) (TokOp "!") >>> expr7 >>> unexpr' "not"
    <|> expr8

unexpr' n = proc e -> do
    returnA -< (builtin n `App` e)

binexpr' n = proc (l,r) -> do
    returnA -< (builtin n `App` l `App` r)

unexpr n = proc e -> do
    returnA -< (global n `AppTy` e)

binexpr n = proc (l,r) -> do
    returnA -< (global n `AppTy` l `App` r)

expr6 = leftassoc [(TokOp "*", binexpr "mul"), (TokOp "/", binexpr "div"),
                   (TokOp "%", binexpr "rem")] expr7

expr5 = leftassoc [(TokOp "+", binexpr "add"), (TokOp "-", binexpr "sub")] expr6

expr4 = nonassoc [(TokOp "<", binexpr "lt"), (TokOp "<=", binexpr "leq"),
                  (TokOp ">", binexpr "gt"), (TokOp ">=", binexpr "geq")] expr5

expr3 = nonassoc [(TokOp "==", binexpr "eq"), (TokOp "!=", binexpr "neq")] expr4

expr2 = leftassoc [(TokOp "&&", binexpr' "and")] expr3

expr1 = leftassoc [(TokOp "||", binexpr' "or")] expr2

expr0 = rightassoc [(TokOp "->", binexpr' "Arrow")] expr1

expr = nonassoc [(TokOp "::", arr (uncurry Ann))] expr0

defs = sepBy1 def ((iskip.tokenP) TokSemicolon)

def = proc () -> do
    (TokName (Attr n)) <- tokenP (TokName AttrPH) -< ()
    tokenP (TokOp "=") -< ()
    e <- expr -< ()
    returnA -< (n, e)

leftassoc l n = n >>> chain (choice $ map (assocH n) l)

rightassoc l n = x where
    x = n >>> choice (map (assocH x) l) <+> emptyP where

nonassoc l n = n >>> choice (map (assocH n) l) <+> emptyP

assocH :: PArrL () Term -> (Token, PArrL (Term, Term) Term) -> PArrL Term Term
assocH r (t,c) = (iskip.tokenP) t >>> emptyP &&& (const () ^>> r) >>> c

parseExpr sn s = case parseAll defs ((),s,initstate sn) of
        Left e -> Left e
        Right (e,_) -> Right e

