module Typecheck(typecheck) where

import Language
import Data.Set(Set)
import qualified Data.Set as Set
import Data.Map(Map)
import qualified Data.Map as Map
import Data.Fixed(mod')
import Control.Monad
import Control.Monad.Error
import Control.Monad.State
import Control.Arrow

import Debug.Trace

type Tc = ErrorT String (State TcState)
type EvEnv = Map Id Term
type TyEnv = Map Id Term
type Ctxt = (EvEnv, TyEnv)
data TcState = TcState {
    snum :: Int,
    evenv :: EvEnv,
    tyenv :: TyEnv
}

genNum :: Tc Int
genNum = do
    s <- get
    put (s { snum = 1 + snum s })
    return (snum s)

evalWhnf :: Term -> Tc Term
evalWhnf t = h [] t where
        h s (App t1 t2) = h (t2:s) t1
        h (a:s) (Lam t) = h s (subst 0 a t)
        h s (Ann t _) = h s t
        h s (Let t1 t2) = h s (subst 0 t1 t2)
        -- casetype
        h s (CaseType t f1 f2 f3 f4 f5 f6 f7 f8 f9 f10) = do
            et <- evalWhnf t
            case et of
                Star -> h s f1
                Free (Builtin "Unit") -> h s f2
                Free (Builtin "Bool") -> h s f3
                Free (Builtin "Int") -> h s f4
                Free (Builtin "Float") -> h s f5
                Free (Builtin "Char") -> h s f6
                App (App (Free (Builtin "Pair")) t1) t2 ->
                    h s (f7 `App` t1 `App` t2)
                App (App (Free (Builtin "Sum")) t1) t2 ->
                    h s (f8 `App` t1 `App` t2)
                Pi t1 t2 -> if Local(-1) `Set.member` freevars(unbound (-1) t2)
                    then h s (f10 `App` t1 `App` (Lam t2))
                    else h s (f9 `App` t1 `App` t2)
        -- builtins
        h s (Free (Builtin "Star")) = h s Star
        h s (Free (Builtin "true")) = h s (BoolV True)
        h s (Free (Builtin "false")) = h s (BoolV False)
        h (l:r:s) (Free (Builtin "Arrow")) = h s (l `Pi` r)
        h (t:f:s) (Free (Builtin "fix")) = 
            h s (f `App` (builtin "fix" `App` t `App` f))
        h (_:_:a:s) (Free (Builtin "wrap")) = h s a
        h (_:_:a:s) (Free (Builtin "unwrap")) = h s a
        h (_:_:a:s) (Free (Builtin "fst")) = do
            ea <- evalWhnf a
            case ea of
                App (App (App (App (Free (Builtin "pair")) _) _) x) y ->
                    h s x
        h (_:_:a:s) (Free (Builtin "snd")) = do
            ea <- evalWhnf a
            case ea of
                App (App (App (App (Free (Builtin "pair")) _) _) x) y -> h s y
        h (_:_:_:a:f:g:s) (Free (Builtin "case")) = do
            ea <- evalWhnf a
            case ea of
                App (App (App (Free (Builtin "inl")) _) _) x -> h s (App f x)
                App (App (App (Free (Builtin "inr")) _) _) y -> h s (App g y)
        h (b:_:at:af:s) (Free (Builtin "if")) = do
            eb <- evalWhnf b
            case eb of
                BoolV True -> h s at
                BoolV False -> h s af
        h (a:s) (Free (Builtin "floattoint")) = do
            (FloatV v) <- evalWhnf a
            h s (IntV (truncate v))
        h (a:s) (Free (Builtin "inttofloat")) = do
            (IntV v) <- evalWhnf a
            h s (FloatV (realToFrac v))
        h (a1:a2:s) (Free (Builtin "addint")) = h_intbin a1 a2 s (+)
        h (a1:a2:s) (Free (Builtin "subint")) = h_intbin a1 a2 s (-)
        h (a1:a2:s) (Free (Builtin "mulint")) = h_intbin a1 a2 s (*)
        h (a1:a2:s) (Free (Builtin "divint")) = h_intbin a1 a2 s div
        h (a1:a2:s) (Free (Builtin "modint")) = h_intbin a1 a2 s mod
        h (a:s) (Free (Builtin "negint")) = h_intun a s negate
        h (a1:a2:s) (Free (Builtin "eqint")) = h_intbbin a1 a2 s (==)
        h (a1:a2:s) (Free (Builtin "neint")) = h_intbbin a1 a2 s (/=)
        h (a1:a2:s) (Free (Builtin "ltint")) = h_intbbin a1 a2 s (<)
        h (a1:a2:s) (Free (Builtin "leint")) = h_intbbin a1 a2 s (<=)
        h (a1:a2:s) (Free (Builtin "addfloat")) = h_floatbin a1 a2 s (+)
        h (a1:a2:s) (Free (Builtin "subfloat")) = h_floatbin a1 a2 s (-)
        h (a1:a2:s) (Free (Builtin "mulfloat")) = h_floatbin a1 a2 s (*)
        h (a1:a2:s) (Free (Builtin "divfloat")) = h_floatbin a1 a2 s (/)
        h (a1:a2:s) (Free (Builtin "modfloat")) = h_floatbin a1 a2 s mod'
        h (a:s) (Free (Builtin "negfloat")) = h_floatun a s negate
        h (a1:a2:s) (Free (Builtin "eqfloat")) = h_floatbbin a1 a2 s (==)
        h (a1:a2:s) (Free (Builtin "nefloat")) = h_floatbbin a1 a2 s (/=)
        h (a1:a2:s) (Free (Builtin "ltfloat")) = h_floatbbin a1 a2 s (<)
        h (a1:a2:s) (Free (Builtin "lefloat")) = h_floatbbin a1 a2 s (<=)
        h (a1:a2:s) (Free (Builtin "eqchar")) = h_charbbin a1 a2 s (==)
        h (a1:a2:s) (Free (Builtin "nechar")) = h_charbbin a1 a2 s (/=)
        h (a1:a2:s) (Free (Builtin "ltchar")) = h_charbbin a1 a2 s (<)
        h (a1:a2:s) (Free (Builtin "lechar")) = h_charbbin a1 a2 s (<=)
        h (a:s) (Free (Builtin "sin")) = h_floatun a s sin
        h (a:s) (Free (Builtin "cos")) = h_floatun a s cos
        h (a:s) (Free (Builtin "tan")) = h_floatun a s tan
        h (a:s) (Free (Builtin "sinh")) = h_floatun a s sinh
        h (a:s) (Free (Builtin "cosh")) = h_floatun a s cosh
        h (a:s) (Free (Builtin "tanh")) = h_floatun a s tanh
        h (a:s) (Free (Builtin "asin")) = h_floatun a s asin
        h (a:s) (Free (Builtin "acos")) = h_floatun a s acos
        h (a:s) (Free (Builtin "atan")) = h_floatun a s atan
        h (a1:a2:s) (Free (Builtin "atan2")) = h_floatbin a1 a2 s atan2
        h (a:s) (Free (Builtin "ceil")) = h_floatun a s (fromIntegral.ceiling)
        h (a:s) (Free (Builtin "floor")) = h_floatun a s (fromIntegral.floor)
        h (a:s) (Free (Builtin "round")) = h_floatun a s (fromIntegral.round)
        h (a:s) (Free (Builtin "abs")) = h_floatun a s abs
        h (a1:a2:s) (Free (Builtin "pow")) = h_floatbin a1 a2 s (**)
        h (a:s) (Free (Builtin "log")) = h_floatun a s log
        h (a:s) (Free (Builtin "log10")) = h_floatun a s (logBase 10)
        h (a:s) (Free (Builtin "sqrt")) = h_floatun a s sqrt
        h (a:s) (Free (Builtin "cbrt")) = h_floatun a s (** (1/3))
        h (a1:a2:s) (Free (Builtin "and")) = h_boolbin a1 a2 s (&&)
        h (a1:a2:s) (Free (Builtin "or")) = h_boolbin a1 a2 s (&&)
        h (a:s) (Free (Builtin "not")) = h_boolun a s not
        h (a1:a2:s) (Free (Builtin "eqv")) = h_boolbin a1 a2 s (==)
        h (a1:a2:s) (Free (Builtin "neqv")) = h_boolbin a1 a2 s (/=)
        h s (Free n) = do
            x <- lookupEvEnv n
            case x of
                Just v -> h s v
                Nothing -> return $ putspine (Free n) s
        h s t = return $ putspine t s
        putspine = foldl App
        h_intbin a1 a2 s o = do
            ea1 <- evalWhnf a1
            ea2 <- evalWhnf a2
            case (ea1, ea2) of
                (IntV v1, IntV v2) -> h s (IntV (o v1 v2))
        h_intbbin a1 a2 s o = do
            ea1 <- evalWhnf a1
            ea2 <- evalWhnf a2
            case (ea1, ea2) of
                (IntV v1, IntV v2) -> h s (BoolV (o v1 v2))
        h_intun a s o = do
            ea <- evalWhnf a
            case ea of
                IntV v -> h s (IntV (o v))
        h_floatbin a1 a2 s o = do
            ea1 <- evalWhnf a1
            ea2 <- evalWhnf a2
            case (ea1, ea2) of
                (FloatV v1, FloatV v2) -> h s (FloatV (o v1 v2))
        h_floatbbin a1 a2 s o = do
            ea1 <- evalWhnf a1
            ea2 <- evalWhnf a2
            case (ea1, ea2) of
                (FloatV v1, FloatV v2) -> h s (BoolV (o v1 v2))
        h_floatun a s o = do
            ea <- evalWhnf a
            case ea of
                FloatV v -> h s (FloatV (o v))
        h_boolbin a1 a2 s o = do
            ea1 <- evalWhnf a1
            ea2 <- evalWhnf a2
            case (ea1, ea2) of
                (BoolV v1, BoolV v2) -> h s (BoolV (o v1 v2))
        h_boolun a s o = do
            ea <- evalWhnf a
            case ea of
                BoolV v -> h s (BoolV (o v)) 
        h_charbbin a1 a2 s o = do
            ea1 <- evalWhnf a1
            ea2 <- evalWhnf a2
            case (ea1, ea2) of
                (CharV v1, CharV v2) -> h s (BoolV (o v1 v2))

termeq :: Int -> Term -> Term -> Tc ()
termeq d t1 t2 = termeqe d t1 t2
        where
        termeqe d t1 t2 = do
            et1 <- evalWhnf t1
            et2 <- evalWhnf t2
            termeq' d et1 et2
        termeq' d (App t1 t2) (App u1 u2) = do
            termeq' d t1 u1 
            termeqe d t2 u2
        termeq' d (Lam t) (Lam u) = do
            sn <- genNum
            termeqe (d+1) (unboundSk sn t) (unboundSk sn u)
        termeq' d (Pi t1 t2) (Pi u1 u2) = do
            sn <- genNum
            termeqe d t1 u1
            termeqe (d+1) (unboundSk sn t2) (unboundSk sn u2)
        termeq' d (Star) (Star) = return ()
        termeq' d (Free u) (Free v) | u == v = return ()
        termeq' _ l r = fail ("unification failed: " ++ show l ++ " " ++ show r)

lookupTyEnv :: Id -> Tc Term
lookupTyEnv i = do
    s <- get
    case Map.lookup i (tyenv s) of
        Just a -> return a
        Nothing -> fail ("lookup fail on " ++ show i)

getEvEnv :: Tc EvEnv
getEvEnv = get >>= (return . evenv)

lookupEvEnv :: Id -> Tc (Maybe Term)
lookupEvEnv i = do
    s <- get 
    return $ Map.lookup i (evenv s)

extendTyEnv :: Id -> Term -> Tc Term -> Tc Term
extendTyEnv i t a = do
    updateE
    r <- a
    retractE
    return r
    where
    -- todo przyslanianie? odtwarzac evenv?
    updateE = do
        s <- get
        put (s{tyenv = Map.insert i t (tyenv s), evenv = Map.delete i (evenv s)})
    retractE = do
        s <- get
        put (s{tyenv = Map.delete i (tyenv s), evenv = Map.delete i (evenv s)})

extendEvEnv :: Id -> Term -> Tc Term -> Tc Term
extendEvEnv i t a = do
    updateE
    (Ann r x) <- a
    retractE
    return (Ann r (substv i t x))
    where
    -- todo odtwarzanie starej wartosci?
    updateE = do
        s <- get
        put (s{evenv = Map.insert i t (evenv s)})
    retractE = do
        s <- get
        put (s{evenv = Map.delete i (evenv s)})

extendTyEvEnv :: Id -> Term -> Term -> Tc Term -> Tc Term
extendTyEvEnv i t v a = extendTyEnv i t $ extendEvEnv i v a

deftyenv = Map.fromList [
    (Builtin "fix", 
        Star `Pi` (Bound 0 `Pi` Bound 1) `Pi` Bound 1),
    (Builtin "Fix", 
        Star `Pi` (Bound 0 `Pi` Bound 1) `Pi` Bound 1),
    (Builtin "wrap",
        Star `Pi` (Bound 0 `Pi` Bound 1) `Pi` 
            (Bound 0 `App` (builtin "Fix" `App` Bound 1 `App` Bound 0)) `Pi`
            (builtin "Fix" `App` Bound 2 `App` Bound 1)),
    (Builtin "unwrap",
        Star `Pi` (Bound 0 `Pi` Bound 1) `Pi` 
            (builtin "Fix" `App` Bound 1 `App` Bound 0) `Pi`
            (Bound 1 `App` (builtin "Fix" `App` Bound 2 `App` Bound 1))),
    (Builtin "Pair", Star `Pi` Star `Pi` Star),
    (Builtin "pair", 
        Star `Pi` Star `Pi` Bound 1 `Pi` Bound 1 `Pi`
            (builtin "Pair" `App` Bound 3 `App` Bound 2)),
    (Builtin "fst", 
        Star `Pi` Star `Pi`
            (builtin "Pair" `App` Bound 1 `App` Bound 0) `Pi` Bound 2),
    (Builtin "snd", 
        Star `Pi` Star `Pi` 
            (builtin "Pair" `App` Bound 1 `App` Bound 0) `Pi` Bound 1),
    (Builtin "Sum", Star `Pi` Star `Pi` Star),
    (Builtin "inl", 
        Star `Pi` Star `Pi` Bound 1 `Pi`
            (builtin "Sum" `App` Bound 2 `App` Bound 1)),
    (Builtin "inr", 
        Star `Pi` Star `Pi` (Bound 0) `Pi`
            (builtin "Sum" `App` Bound 2 `App` Bound 1)),
    (Builtin "case", 
        Star `Pi` Star `Pi` Star `Pi`
            (builtin "Sum" `App` Bound 2 `App` Bound 1) `Pi`
                (Bound 3 `Pi` Bound 2) `Pi` (Bound 3 `Pi` Bound 3) `Pi`
                    Bound 3),
    (Builtin "Int", Star),
    (Builtin "addint", 
        builtin "Int" `Pi` builtin "Int" `Pi` builtin "Int"),
    (Builtin "subint", 
        builtin "Int" `Pi` builtin "Int" `Pi` builtin "Int"),
    (Builtin "mulint", 
        builtin "Int" `Pi` builtin "Int" `Pi` builtin "Int"),
    (Builtin "divint", 
        builtin "Int" `Pi` builtin "Int" `Pi` builtin "Int"),
    (Builtin "remint", 
        builtin "Int" `Pi` builtin "Int" `Pi` builtin "Int"),
    (Builtin "negint", builtin "Int" `Pi` builtin "Int"),
    (Builtin "eqint", 
        builtin "Int" `Pi` builtin "Int" `Pi` builtin "Bool"),
    (Builtin "neint", 
        builtin "Int" `Pi` builtin "Int" `Pi` builtin "Bool"),
    (Builtin "ltint", 
        builtin "Int" `Pi` builtin "Int" `Pi` builtin "Bool"),
    (Builtin "leint", 
        builtin "Int" `Pi` builtin "Int" `Pi` builtin "Bool"),
    (Builtin "Char", Star),
    (Builtin "Bool", Star),
    (Builtin "true", builtin "Bool"),
    (Builtin "false", builtin "Bool"),
    (Builtin "and", 
        builtin "Bool" `Pi` builtin "Bool" `Pi` builtin "Bool"),
    (Builtin "or", 
        builtin "Bool" `Pi` builtin "Bool" `Pi` builtin "Bool"),
    (Builtin "eqv", 
        builtin "Bool" `Pi` builtin "Bool" `Pi` builtin "Bool"),
    (Builtin "neqv", 
        builtin "Bool" `Pi` builtin "Bool" `Pi` builtin "Bool"),
    (Builtin "not", Pi (builtin "Bool") (builtin "Bool")),
    (Builtin "if", 
        builtin "Bool" `Pi` Star `Pi` Bound 0 `Pi` Bound 1 `Pi` Bound 2),
    (Builtin "Float", Star),
    (Builtin "addfloat", 
        builtin "Float" `Pi` builtin "Float" `Pi` builtin "Float"),
    (Builtin "subfloat", 
        builtin "Float" `Pi` builtin "Float" `Pi` builtin "Float"),
    (Builtin "mulfloat", 
        builtin "Float" `Pi` builtin "Float" `Pi` builtin "Float"),
    (Builtin "divfloat", 
        builtin "Float" `Pi` builtin "Float" `Pi` builtin "Float"),
    (Builtin "remfloat", 
        builtin "Float" `Pi` builtin "Float" `Pi` builtin "Float"),
    (Builtin "negfloat", builtin "Float" `Pi` builtin "Float"),
    (Builtin "eqfloat", 
        builtin "Float" `Pi` builtin "Float" `Pi` builtin "Bool"),
    (Builtin "nefloat", 
        builtin "Float" `Pi` builtin "Float" `Pi` builtin "Bool"),
    (Builtin "ltfloat", 
        builtin "Float" `Pi` builtin "Float" `Pi` builtin "Bool"),
    (Builtin "lefloat", 
        builtin "Float" `Pi` builtin "Float" `Pi` builtin "Bool"),
    (Builtin "floattoint", builtin "Float" `Pi` builtin "Int"),
    (Builtin "inttofloat", builtin "Int" `Pi` builtin "Float"),
    (Builtin "sin", builtin "Float" `Pi` builtin "Float"),
    (Builtin "cos", builtin "Float" `Pi` builtin "Float"),
    (Builtin "tan", builtin "Float" `Pi` builtin "Float"),
    (Builtin "sinh", builtin "Float" `Pi` builtin "Float"),
    (Builtin "cosh", builtin "Float" `Pi` builtin "Float"),
    (Builtin "tanh", builtin "Float" `Pi` builtin "Float"),
    (Builtin "asin", builtin "Float" `Pi` builtin "Float"),
    (Builtin "acos", builtin "Float" `Pi` builtin "Float"),
    (Builtin "atan", builtin "Float" `Pi` builtin "Float"),
    (Builtin "atan2", 
        builtin "Float" `Pi` builtin "Float" `Pi` builtin "Float"),
    (Builtin "ceil", builtin "Float" `Pi` builtin "Float"),
    (Builtin "floor", builtin "Float" `Pi` builtin "Float"),
    (Builtin "round", builtin "Float" `Pi` builtin "Float"),
    (Builtin "abs", builtin "Float" `Pi` builtin "Float"),
    (Builtin "pow", builtin "Float" `Pi` builtin "Float" `Pi` builtin "Float"),
    (Builtin "sqrt", builtin "Float" `Pi` builtin "Float"), 
    (Builtin "cbrt", builtin "Float" `Pi` builtin "Float"), 
    (Builtin "log", builtin "Float" `Pi` builtin "Float"), 
    (Builtin "log10", builtin "Float" `Pi` builtin "Float"), 
    (Builtin "eqchar", 
        builtin "Char" `Pi` builtin "Char" `Pi` builtin "Bool"),
    (Builtin "nechar", 
        builtin "Char" `Pi` builtin "Char" `Pi` builtin "Bool"),
    (Builtin "ltchar", 
        builtin "Char" `Pi` builtin "Char" `Pi` builtin "Bool"),
    (Builtin "lechar", 
        builtin "Char" `Pi` builtin "Char" `Pi` builtin "Bool"),
    (Builtin "Unit", Star),
    (Builtin "unit", builtin "Unit"),
    (Builtin "Star", Star),
    (Builtin "Arrow", Star `Pi` Star `Pi` Star),
    (Builtin "IO", Star `Pi` Star),
    (Builtin "return", Star `Pi` Bound 0 `Pi` (builtin "IO" `App` Bound 1)),
    (Builtin "bind", Star `Pi` Star `Pi` (builtin "IO" `App` Bound 1) `Pi`
        (Bound 2 `Pi` (builtin "IO" `App` Bound 2)) `Pi`
            (builtin "IO" `App` Bound 2)),
    (Builtin "readint", builtin "IO" `App` builtin "Int"),
    (Builtin "readfloat", builtin "IO" `App` builtin "Float"),
    (Builtin "readchar", builtin "IO" `App` builtin "Char"),
    (Builtin "printint", 
        builtin "Int" `Pi` (builtin "IO" `App` builtin "Unit")),
    (Builtin "printfloat", 
        builtin "Float" `Pi` (builtin "IO" `App` builtin "Unit")),
    (Builtin "printchar", 
        builtin "Char" `Pi` (builtin "IO" `App` builtin "Unit"))
    ]

initstate = TcState 0 Map.empty deftyenv

termEq :: Int -> Term -> Term -> Tc ()
termEq d t1 t2 = termeq d t1 t2

typeIn :: Int -> Term -> Tc Term
typeIn a b = {- trace ("typeIn " ++ show a ++ " " ++ show b ++ " ") -} (typeIn' a b)

typeIn' :: Int -> Term -> Tc Term
typeIn' d (Ann e t) = do
    t' <- typeCh d t Star
    t'' <- evalWhnf t'
    e' <- typeCh d e t''
    return (Ann e' t')
typeIn' d Star = return (Ann Star Star)
typeIn' d (Pi t1 t2) = do
    t1' <- typeCh d t1 Star
    t1'' <- evalWhnf t1'
    t2' <- extendTyEnv (Local d) t1'' $ typeCh (d+1) (unbound d t2) Star
    return (Ann (Pi t1 (bound d t2')) Star)
typeIn' d (Free x) = do
    v <- lookupTyEnv x
    t <- evalWhnf v
    return (Ann (Free x) t)
typeIn' d (Let e1 e2) = do
    (Ann e1' t1) <- typeIn d e1
    (Ann e2' t) <- extendTyEvEnv (Local d) t1 e1' $ typeIn (d+1) (unbound d e2)
    return (Ann (Let e1' (bound d e2')) t)
typeIn' d (IntV v) = return (Ann (IntV v) (builtin "Int"))
typeIn' d (BoolV v) = return (Ann (BoolV v) (builtin "Bool"))
typeIn' d (FloatV v) = return (Ann (FloatV v) (builtin "Float"))
typeIn' d (CharV v) = return (Ann (CharV v) (builtin "Char"))
typeIn' d (App e1 e2) = do
    (Ann e1' t1) <- typeIn d e1
    case t1 of
        Pi t2 t3 -> do
            t2' <- evalWhnf t2
            e2' <- typeCh d e2 t2'
            t3' <- evalWhnf (subst 0 e2 t3)
            return (Ann (App e1' e2') t3')
        t -> do
            fail ("illegal application " ++ show t ++ " " ++ show e1 ++ " ; " ++ show (App e1 e2))
typeIn' d (AppTy e1 e2) = do
    (Ann e2' t2) <- typeIn d e2
    typeIn d (App (App e1 t2) e2')
typeIn' d e = fail ("lol " ++ show e)

typeCh :: Int -> Term -> Term -> Tc Term
typeCh a b c = {- trace ("typeCh " ++ show a ++ " " ++ show b ++ " " ++ show c) -} (typeCh' a b c)

typeCh' :: Int -> Term -> Term -> Tc Term
typeCh' d (Lam e) (Pi t1 t2) = do
    t2' <- evalWhnf (unbound d t2)
    e' <- extendTyEnv (Local d) t1 $ typeCh (d+1) (unbound d e) t2'
    return (Ann (Lam (bound d e')) (Pi t1 t2'))
typeCh' d (CaseType (Free i) f1 f2 f3 f4 f5 f6 
    (Lam (Lam f7)) (Lam (Lam f8)) (Lam (Lam f9)) (Lam (Lam f10))) t1 = do
    evenv <- getEvEnv
    typeCh d (Free i) Star
    when (i `Map.member` evenv) $ fail "casetype on defined type"
    f1' <- extendEvEnv i Star $ typeCh d f1 t1
    f2' <- extendEvEnv i (builtin "Unit") $ typeCh d f2 t1
    f3' <- extendEvEnv i (builtin "Bool") $ typeCh d f3 t1
    f4' <- extendEvEnv i (builtin "Int") $ typeCh d f4 t1
    f5' <- extendEvEnv i (builtin "Float") $ typeCh d f5 t1
    f6' <- extendEvEnv i (builtin "Char") $ typeCh d f6 t1
    f7' <- extendTyEnv (Local d) Star $ extendTyEnv (Local (d+1)) Star $
        extendEvEnv i (builtin "Pair" `App` local d `App` local (d+1)) $ 
             typeCh (d+2) (unbound (d+1) $ subst 1 (local d) f7) t1
    f8' <- extendTyEnv (Local d) Star $ extendTyEnv (Local (d+1)) Star $
        extendEvEnv i (builtin "Sum" `App` local d `App` local (d+1)) $ 
            typeCh (d+2) (unbound (d+1) $ subst 1 (local d) f8) t1
    f9' <- extendTyEnv (Local d) Star $ extendTyEnv (Local (d+1)) Star $
        extendEvEnv i (local d `Pi` local (d+1)) $ 
            typeCh (d+2) (unbound (d+1) $ subst 1 (local d) f9) t1
    f10' <- extendTyEnv (Local d) Star $ 
        extendTyEnv (Local (d+1)) (local d `Pi` Star) $
            extendEvEnv i (local d `Pi` (local (d+1) `App` Bound 0)) $ 
                typeCh (d+2) (unbound (d+1) $ subst 1 (local d) f10) t1
    return (Ann (CaseType (Free i) f1' f2' f3' f4' f5' f6' (Lam (Lam f7')) (Lam (Lam f8')) (Lam (Lam f9')) (Lam (Lam f10'))) t1)
typeCh' d e ct = do
    (Ann e' t) <- typeIn d e
    termEq d t ct
    return (Ann e' t)

typeChAll t ty = typeCh 0 t ty

typecheck t = evalState (runErrorT $ typeChAll t (builtin "IO" `App` builtin "Unit")) initstate

